from django.contrib.auth import get_user_model, backends
from django.conf import settings

from sssauth.settings import app_settings
from sssauth.utils import recover_to_addr
from sssauth.utils import publicKey_to_addr


class Web3Backend(backends.ModelBackend):
    def authenticate(self, request, signature=None):
        # get user model
        User = get_user_model()
        # check if the address the user has provided matches the signature
        if not recover_to_addr(signature):
            return None
        else:
            print("validate success")
            # get address field for the user model
            address_field = app_settings.SSSAUTH_USER_ADDRESS_FIELD
            kwargs = {
                f"{address_field}__iexact": publicKey_to_addr(signature[:64],settings.NETWORK_TYPE)
            }
            # try to get user with provided data
            user = User.objects.filter(**kwargs).first()
            return user



from django.urls import path

from sssauth import views

app_name = 'sssauth'

urlpatterns = [
    path('login_api/', views.login_api, name='sssauth_login_api'),
    path('signup_api/', views.signup_api, name='sssauth_signup_api'),
]




# Generated by Django 4.0 on 2022-08-29 05:16

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sssauth', '0003_alter_myuser_username'),
    ]

    operations = [
        migrations.AlterField(
            model_name='myuser',
            name='username',
            field=models.CharField(help_text='Required. Address in the format containing "-" are also allowed.', max_length=45, unique=True, validators=[django.core.validators.MinLengthValidator(39), django.core.validators.RegexValidator('^[a-zA-Z0-9]*$')], verbose_name='address'),
        ),
    ]



import string

from django import forms
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _

from sssauth.settings import app_settings
from sssauth.utils import validate_xym_address

CustomUser = get_user_model()

class LoginForm(forms.Form):
    payload = forms.CharField(widget=forms.HiddenInput, min_length=64)

    def __init__(self, token, *args, **kwargs):
        self.token = token
        super(LoginForm, self).__init__(*args, **kwargs)

    def clean_signature(self):
        sig = self.cleaned_data['payload']
        if sig == "":
            raise forms.ValidationError(_('Invalid signature'))
        return sig


# list(set()) here is to eliminate the possibility of double including the address field
signup_fields = list(set(app_settings.SSSAUTH_USER_SIGNUP_FIELDS + [app_settings.SSSAUTH_USER_ADDRESS_FIELD]))


class SignupForm(forms.ModelForm):

    def __init__(self, *args, **kwargs):
        # first call parent's constructor
        super().__init__(*args, **kwargs)

        # make sure to make email required, because password is not set
        # and if the user loses private key he can get 'reset' password link to email
        if 'email' in app_settings.SSSAUTH_USER_SIGNUP_FIELDS:
            self.fields['email'].required = True
        self.fields[app_settings.SSSAUTH_USER_ADDRESS_FIELD].required = True

    def clean_address_field(self):
        addr = self.cleaned_data[app_settings.SSSAUTH_USER_ADDRESS_FIELD].replace("-","")
        validate_xym_address(addr)
        return addr

    class Meta:
        model = CustomUser
        fields = signup_fields


# hack to set the method for cleaning address field
setattr(SignupForm, 'clean_' + app_settings.SSSAUTH_USER_ADDRESS_FIELD, SignupForm.clean_address_field)



from django.db import models

from django.contrib.auth.models import BaseUserManager, AbstractBaseUser, PermissionsMixin
from django.core.validators import MinLengthValidator, RegexValidator

class MyUserManager(BaseUserManager):
    def create_user(self, address, email):
        if not address:
            raise ValueError('Users must have an Address')
        if not email:
            raise ValueError('Users must have an email address')

        user = self.model(
            address=address,
            email=self.normalize_email(email),
        )
        user.save(using=self._db)
        return user

    def create_superuser(self, address, email):
        user = self.create_user(
            address=address,
            email=self.normalize_email(email),
        )
        user.is_admin=True
        user.is_staff=True
        user.is_superuser=True
        user.save(using=self._db)
        return user

class MyUser(AbstractBaseUser, PermissionsMixin):

    address = models.CharField(verbose_name='address', max_length=45, unique=True, validators=[MinLengthValidator(39,), RegexValidator(r'^[a-zA-Z0-9]*$',)])
    email = models.EmailField(verbose_name='email', max_length=50, unique=True)
    date_joined = models.DateTimeField(verbose_name='joined', auto_now_add=True)
    is_active = models.BooleanField(default=True)
    is_staff = models.BooleanField(default=False)
    is_admin = models.BooleanField(default=False)

    objects = MyUserManager()
    # identify
    USERNAME_FIELD = 'address'
    # list
    REQUIRED_FIELDS = ['email']

    def __str__(self):
        return self.address


import json
import symbolhkdf
import time
import binascii
import base64
import hashlib

from django import forms
from django.utils.translation import gettext_lazy as _
from django.conf import settings

def publicKey_to_addr(publicKey, networkType = 152):
    if len(publicKey)!= 64:
        raise ValueError("Invalid publicKey")
    part_one_hash_builder = hashlib.sha3_256()
    part_one_hash_builder.update(binascii.unhexlify(publicKey))
    part_one_hash = part_one_hash_builder.digest()

    part_two_hash_builder = hashlib.new('ripemd160')
    part_two_hash_builder.update(part_one_hash)
    part_two_hash = part_two_hash_builder.digest()

    version = bytes([networkType]) + part_two_hash

    part_three_hash_builder = hashlib.sha3_256()
    part_three_hash_builder.update(version)
    checksum = part_three_hash_builder.digest()[0:3]
    address_encode = base64.b32encode(version+checksum).decode('utf-8')[:-1]
    return address_encode

def recover_to_addr(payload):
    sender_publickey = payload[:64]
    encrypted_message = payload[64:]
    decode = json.loads(symbolhkdf.decode(settings.SERVER_SECRET, sender_publickey, encrypted_message))#復号
    print("addr:", publicKey_to_addr(sender_publickey,settings.NETWORK_TYPE))
    iat = decode["iat"] 
    signer = decode["signerAddress"]
    verifier = decode["verifierAddress"]
    if verifier != settings.OWNER:
        raise ValueError("Invalid verifier")
    if time.time()*1000 - int(iat) - settings.EXPIRATION_DATE > 0:
        raise ValueError("iat expired")
    return signer


def validate_xym_address(value):# サインアップ時のアドレス検証
    if not len(value) == 39:
        raise forms.ValidationError(
            _('%(value)s is not a valid Symbol address'),
            params={'value': value},
        )
    decode = base64.b32decode(value+"=")
    checksum = decode[21:]
    hasher = hashlib.sha3_256()
    calculate_chk = hasher.update(decode[:21])
    calculate_chk = hasher.digest()[:len(checksum)]
    print("decode:",decode,"chk:",checksum,"calc:",calculate_chk)
    if checksum != calculate_chk:
        raise forms.ValidationError(
            _('%(value)s is not a valid Symbol address'),
            params={'value': value},
        )




from django.conf import settings as django_settings


class AppSettings(object):
    @property
    def SSSAUTH_USER_ADDRESS_FIELD(self):
        """
        Field on the User model, which has catapult address to check against.
        This allows you to store it somewhere in arbitrary place other than just username.
        """
        return getattr(django_settings, 'SSSAUTH_USER_ADDRESS_FIELD', 'address')

    @property
    def SSSAUTH_USER_SIGNUP_FIELDS(self):
        """
        Specifies field to be used in signup form for a new User model
        """
        return getattr(django_settings, "SSSAUTH_USER_SIGNUP_FIELDS", ['email'])

    @property
    def SSSAUTH_SIGNUP_ENABLED(self):
        """
        Makes it possible to disable signups (similar to allauth)
        """
        return getattr(django_settings, "SSSAUTH_SIGNUP_ENABLED", True)


app_settings = AppSettings()



import json

from django.conf import settings
from django.contrib.auth import login, authenticate
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import JsonResponse
from django.shortcuts import render, redirect, reverse
from django.urls.exceptions import NoReverseMatch
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_http_methods
from django.views.generic import TemplateView

from sssauth.forms import LoginForm, SignupForm
from sssauth.settings import app_settings


def get_redirect_url(request):
    if request.GET.get('next'):
        return request.GET.get('next')
    elif request.POST.get('next'):
        return request.POST.get('next')
    elif settings.LOGIN_REDIRECT_URL:
        try:
            url = reverse(settings.LOGIN_REDIRECT_URL)
        except NoReverseMatch:
            url = settings.LOGIN_REDIRECT_URL
        return url


@require_http_methods(["GET", "POST"])
def login_api(request):
    if request.method == 'GET':
        token = settings.PUB
        request.session['login_token'] = token
        return JsonResponse({'data': token, 'success': True})
    else:
        print("auth...")
        token = request.session.get('login_token')
        print("token",token)
        print("form:",request.POST)
        if not token:
            return JsonResponse({'error': _(
                "No login token in session, please request token again by sending GET request to this url"),
                'success': False})
        else:
            form = LoginForm(token, request.POST)
            if form.is_valid():
                payload = form.cleaned_data.get("payload")
                # signature, address = form.cleaned_data.get("payload"), form.cleaned_data.get("address")
                print("payload:",payload)
                del request.session['login_token']
                try:
                    user = authenticate(request, signature=payload)
                    print(user)
                    if user:
                        login(request, user, 'sssauth.backend.Web3Backend')

                        return JsonResponse({'success': True, 'redirect_url': get_redirect_url(request)})
                    else:
                        error = _("Can't find a user for the provided signature with address {address}").format(
                            address=payload[:64])
                        return JsonResponse({'success': False, 'error': error})
                except ValueError:
                    return JsonResponse({'success': False, 'error': "invalid payload"})
            else:
                print("form error")
                return JsonResponse({'success': False, 'error': json.loads(form.errors.as_json())})


@require_http_methods(["POST"])
def signup_api(request):
    if not app_settings.SSSAUTH_SIGNUP_ENABLED:
        return JsonResponse({'success': False, 'error': _("Sorry, signup's are currently disabled")})
    form = SignupForm(request.POST)
    if form.is_valid():
        user = form.save(commit=False)
        addr_field = app_settings.SSSAUTH_USER_ADDRESS_FIELD
        setattr(user, addr_field, form.cleaned_data[addr_field])
        user.save()
        login(request, user, 'sssauth.backend.Web3Backend')
        return JsonResponse({'success': True, 'redirect_url': get_redirect_url(request)})
    else:
        return JsonResponse({'success': False, 'error': json.loads(form.errors.as_json())})


@require_http_methods(["GET", "POST"])
def signup_view(request, template_name='sssauth/signup.html'):
    """
    1. Creates an instance of a SignupForm.
    2. Checks if the registration is enabled.
    3. If the registration is closed or form has errors, returns form with errors
    4. If the form is valid, saves the user without saving to DB
    5. Sets the user address from the form, saves it to DB
    6. Logins the user using sssauth.backend.Web3Backend
    7. Redirects the user to LOGIN_REDIRECT_URL or 'next' in get or post params
    :param request: Django request
    :param template_name: Template to render
    :return: rendered template with form
    """
    form = SignupForm()
    if not app_settings.SSSAUTH_SIGNUP_ENABLED:
        form.add_error(None, _("Sorry, signup's are currently disabled"))
    else:
        if request.method == 'POST':
            form = SignupForm(request.POST)
            if form.is_valid():
                user = form.save(commit=False)
                addr_field = app_settings.SSSAUTH_USER_ADDRESS_FIELD
                setattr(user, addr_field, form.cleaned_data[addr_field])
                user.save()
                login(request, user, 'sssauth.backend.Web3Backend')
                return redirect(get_redirect_url(request))
    return render(request,
                  template_name,
                  {'form': form})
